; Programa: Package para entorno para agente de juego a Sokoban 
;           ver http://e-contento.com/sokoban/
;
; Asignatura: Introduccin a la Inteligencia Artificial
;             ETSII - ULL
; Prctica: 5a (sistemas de produccin)
; Profesor: Patricio Garca (patricio@etsii.ull.es)
; Fecha: 30/5/2005
; Comentarios:
;           Para representar el laberinto se utilizan estructuras
;           tipo soko (ver (help 'soko)).
;           Las funciones accesibles mas importantes son:
;	       inic-soko : Comienza a ejecutar el agente
;              gen-soko : Devuelve estructura con estado de inicio
;              vis-soko : Visualiza estructura tipo soko
;              mov-val-soko: Chequea si un movimiento es vlido para
;                            un estado
;              mov-soko: Modifica un estado aplicando un movimiento
;              copy-est-soko: Genera una copia de un estado
;              sol-val-soko: Chequea si un estado es estado final
;              version-soko: Devuelve el nmero de versin
;
;           Ver el correspondiente help para cada una de ellas.
;
;           Para poder utilizarse este paquete ha de cargarse mediante:
;              (load "sokoban")
;              (use-package "sokoban")
;
; Modificaciones:
;           - Ver 0.8: Traslado de ciertas funciones a check-sk.lsp
;                      Incorporacin de cargador de check-sk.o
;                      Algunos cambios de comparar con bloques.lsp
;
;           - Ver 0.7: Primera versin ao 2005, cambios de maquillaje
;             nicamente
;
; Por hacer:
;           - Poner mas comentarios correctos
;           - Probar bajo otros lisps
;           - Chekeos de consistencia de datos de entrada
;

(make-package "sokoban" :nicknames '(sk sokoban) :use '("LISP"))
(in-package "sokoban")

(export '(version-soko *verbose-soko* *blq-soko* *obs-soko* soko make-soko soko-mat soko-robot soko-dest copy-est-soko gen-soko mov-val-soko mov-soko vis-soko sol-val-soko inic-soko))


; Variables globales
; ------------------

(defvar *version-soko* 0.8 "Version de sokoban.lsp")

(defun version-soko ()
  " - Funcin: VERSION-SOKO ()
     Package:SOKOBAN

Imprime la versin del paquete sokoban que se ha cargado"
  *version-soko*)

(defvar *verbose-soko* nil
  "- Variable: TIPO booleano

Activador de visualizacin traza de depuracin. T=activo, NIL=inactivo")


; Estructura de representacin del problema a resolver
; ----------------------------------------------------

(defvar *blq-soko* 'blq
  " - Variable: TIPO simbolo 

Contiene el smbolo que representa los bloques ($) dentro de la matiz del campo MAT de la estructura tipo SOKOBAN.")

(defvar *obs-soko* 'obs
  " - Variable: TIPO smbolo 

Contiene el smbolo que representa los obstculos (#) dentro de la matiz del campo MAT de la estructura tipo SOKOBAN.")

(defstruct soko
  " - Estructura: SOKO (DEFSTRUCT MAT ROBOT DEST)
     Package:SOKOBAN

Estructura que almacena un laberinto Sokoban, junto a las posiciones del robot, los bloque y los destinos. Consta de tres campos: MAT es una matriz de dos dimensiones, ROBOT y DEST son listas.

La matriz MAT representa el laberinto y los bloque a mover, y tiene por lo tanto las dimensiones mximas de ste. Las paredes del laberinto (#) se representa con el smbolo contenido en la variable *OBS-SOKO*, los bloques ($) se representan con el smbolo contenido en la variable *BLQ-SOKO* (ver (help '*OBS-SOKO*) y (help '*BLQ-SOKO*))'. Los espacios vacios estan representados mediante NIL.

El campo ROBOT almacena una lista con dos nmeros, indicando la fila y columna donde esta el robot (@) actualmente.

El campo DEST guarda la informacin de los destinos (.) donde se deberan colocar los bloques, consta de una lista con sublistas de dos elemento indicando fila y columna del destino, contendr tantas sublistas como bloques existan en MAT.

La mayora de funciones de este paquete hace uso de este tipo de estructura."
  mat
  robot
  dest)

; Funciones para creacion del problema a resolver
; -----------------------------------------------

(defun trcar (car)
  (case car
    (#\# *obs-soko*)
    (#\$ *blq-soko*)
    (#\@ 'robot)
    (#\. 'dest)
    (#\* 'blq-dest)
    (#\+ 'rob-dest)
    (otherwise nil)))

(defun line2list (line)
  (if (stringp line)
      (let (list)
	(dotimes (i (length line) (reverse list))
	  (setf list (cons (trcar (aref line i)) list))))
    line))

(defun elim-robot-dest (elm)
  (case elm
    ((robot dest rob-dest) nil)
    (blq-dest *blq-soko*)
    (otherwise elm)))

(defun cribalist (list)
  (mapcar #'elim-robot-dest list))

(defun busca-robot (robot fil list)
  (if (not robot)
      (if (member 'robot list)
	  (list fil (position 'robot list))
	(if (member 'rob-dest list)
	    (list fil (position 'rob-dest list))))
    robot))

(defun es-dest (elm)
  (member elm '(dest blq-dest rob-dest)))

(defun busca-dest (dest fil col list)
  (let* ((tipo  (find-if #'es-dest list))
	 (colrel (if tipo (position tipo list))))
    (if colrel
	(cons (list fil (+ col colrel))
	      (busca-dest dest fil (+ col colrel 1)
			  (rest (subseq list colrel))))
       dest)))

(defun listnnil (n)
  (if (zerop n)
      nil
    (cons nil (listnnil (1- n)))))

(defun completa (cols list)
  (dolist (sublis list list)
    (nconc sublis (listnnil (- cols (length sublis))))))

(defun crea-est (listinic robot-def dest-def maxfil maxcol)
  (setf listinic (completa maxcol listinic))
  (let ((est-ini (make-soko)))
    (setf (soko-mat est-ini) (make-array 
			      (list maxfil maxcol)
			      :initial-contents listinic))
    (setf (soko-robot est-ini) robot-def)
    (setf (soko-dest est-ini) dest-def)
    est-ini))


(defun load-lab-soko (infile)
  (do* ((list (line2list (read-line infile nil 'eof)) 
	      (line2list (read-line infile nil 'eof)))
	(fil 0)
	(maxcol 0)
	(robot nil)
	(dest  nil)
	(listmat nil))
      ((eq list 'eof) (crea-est (reverse listmat) 
				robot dest fil maxcol))
    (setf robot (busca-robot robot fil list))
    (setf dest (busca-dest dest fil 0 list))
    (setf fil (1+ fil))
    (setf maxcol (max maxcol (length list)))
    (setf listmat (cons (cribalist list) listmat))))

(defun load-lab-fich-soko (fichero)
  (with-open-file (infile fichero)
		  (load-lab-soko infile)))

(defun load-lab-str-soko (string)
  (let ((infile (make-string-input-stream string)))
    (load-lab-soko infile)))


; Problemas iniciales internos
; ----------------------------

(defvar str-def 
"  #### 
###  ##
#     #
# #.#@#
# #   #
#   $ #
##   ##
 ##### "
  "String con laberinto defecto")

(defvar robot-def1 '(6 4)
  "Posicin del robot en el laberinto 1 por defecto")

(defvar dest-def1 '((4 4))
  "Lista de destinos del laberinto 1 por defecto")

(defvar robot-def2 '(6 3)
  "Posicin del robot en el laberinto 2 por defecto")

(defvar robot-def3 '(5 3)
  "Posicin del robot en el laberinto 3 por defecto")

(defvar def-n (make-array '(5 2)
			  :initial-contents
			  '(((3 5) ((3 3)))
			    ((6 4) ((4 4)))
			    ((6 3) ((4 4)))
			    ((5 3) ((4 4)))
			    ((5 2) ((4 4)))))
  "Valores robot y dest para laberintos internos")

(defun gen-soko-def (n)
  "Genera laberinto por defecto y/o internos"
  (let ((est-ini (load-lab-str-soko str-def)))
    (if (and (numberp n) (< n (array-dimension def-n 0)))
	(progn
	  (setf (soko-robot est-ini) 
		(copy-list (aref def-n n 0)))
	  (setf (soko-dest est-ini) 
		(copy-alist (aref def-n n 1)))))
    est-ini))

	    
(defun copy-array (arr)
  (let ((arr2 (make-array `(,(array-dimension arr 0)
		,(array-dimension arr 1)))))
    (dotimes (i (array-dimension arr 0) arr2)
      (dotimes (j (array-dimension arr 1))
	(setf (aref arr2 i j) (aref arr i j))))))

(defun copy-est-soko (est)
  " - Funcin: COPY-EST-SOKO (EST)
     Package:SOKOBAN

Copia el estado EST a otras posiciones de memoria, devolviendo dicha copia. EST ha de ser una estructura tipo SOKO. Utilizar para evitar efectos de borde al modificar estados."
  (let ((nue-est (make-soko)))
    (setf (soko-mat nue-est) 
	  (copy-array (soko-mat est)))
    (setf (soko-robot nue-est) 
	  (copy-list (soko-robot est)))
    (setf (soko-dest nue-est)
	  (copy-tree (soko-dest est)))
    nue-est))

  
; Generacion estructura inicial
; -----------------------------

(defvar fichero nil
  "Fichero o numero de laberinto a cargar")

(defvar save-est-ini nil
  "Estado inicial salvaguardado")

(defun gen-soko ()
  " - Funcin: GEN-SOKO (EST)
     Package:SOKOBAN

Genera un estado inicial con un laberinto para el Sokoban, de acuerdo a las especificiaciones del argumento TIPO pasado a la funcin INIC-SOKO, puede cargar el laberinto desde un fichero si asi se le ha especificado en esta variable (ver (help 'INIC-SOKO). Devuelve una estructura tipo SOKO."
  (let ((est-ini
	 (if (or (null fichero) (numberp fichero))
	     (gen-soko-def fichero)
	   (load-lab-fich-soko fichero))))
    ;Salvaguarda el estado inicial en  save-est-ini
    (if est-ini
      (setf save-est-ini (copy-est-soko est-ini)))
    est-ini))


; Visualizacin de estructura de problema
; ---------------------------------------

(defun vis-soko (est)
  " - Funcin: VIS-SOKO (EST)
     Package:SOKOBAN

Visualiza el estado EST utlizando caracteres: @ para el robot, # para los objetos (paredes), $ para los bloques a mover, '.' para los destinos, '*' cuando hay un bloque sobre un destino y '+' cuando el robot esta en un destino. EST ha de ser una estructura tipo SOKO."
  (dotimes (i (array-dimension (soko-mat est) 0) t)
    (terpri)
    (dotimes (j (array-dimension (soko-mat est) 1) t)
      (case (aref (soko-mat est) i j)
	(obs (princ #\#))
	(blq (if (member `(,i ,j) (soko-dest est) :test 'equal)
		 (princ #\*)
		 (princ #\$)))
	(otherwise (cond 
		    ((member `(,i ,j) (soko-dest est) :test 'equal)
		     (if (equal `(,i ,j) (soko-robot est))
			 (princ #\+)
			 (princ #\.)))
		    ((equal `(,i ,j) (soko-robot est))
		     (princ #\@))
		    (t (princ " "))))))))


; Cargamos Chequeador de soluciones
; ---------------------------------

(if (directory "check-sk.o")
    (load "check-sk")        ; Disponible en fichero check-bl.o
  (progn
    (format t "~%Chequeador de soluciones y opcin HOF no disponibles, necesita descargar fichero check-sk.o~%")
    (load "check-sk-null"))) ; Sino se carga el suplente


; Lanzador de agentes y problemas
; -------------------------------

(defun getlisttip (tipo)
  "Obtiene lista de tipos a partir de tipo"
  (cond
   ((stringp tipo) (directory tipo))
   ((numberp tipo) (list tipo))
   ((null tipo) (list tipo))
   ((listp tipo) (apply #'nconc (mapcar #'getlisttip tipo)))
   (t '(()))))


; Comienza a ejecutar el agente pasado
; ------------------------------------
; Parmetros: 
;
; Variables globales afectadas:
;
; Valor de retorno:
;             T    Todo funciono correctamente
;             NIL Hubo algn problema
;
; Requerimientos:
;
; Ejemplo llamada: (INIC-SOKO #'agente-humano :tipo 2) -> (1 (2 1) 1)

(defun inic-soko (&optional
			 (agente #'agente-humano) ;jugador
			 ; debug        ; modo debug
			 &key 
			 verb         ; modo verbose
			 (num nil)     ; Numero de partidas a jugar
			 (tipo nil)   ; Tipo de partidas
			 hof)         ; Aadir resultado al hof
  
  " - Funcin: INIC-SOKO (&OPTIONAL (AGENTE #'agente-humano) &KEY VERB (NUM 1) (TIPO NIL) HOF)
     Package:SOKOBAN

Inicia la ejecucion del agente AGENTE que juega al Sokoban. La funcion AGENTE debe encontrar una solucion y devolverla como una lista de movimientos, cada movimiento puede ser 1, 2, 3  4 segun sean norte, sur, este u oeste. VERB indica si se activa el modo verbose. NUM indica el nmero de partidas a jugar. TIPO indica el tipo de juego, puede ser NIL, 1, 2, 3, 4, una lista de estos o bien un string que cargar el laberinto desde el/los ficheros que empareje. HOF indica si se graban los resultados en el Hall-of-Fame (/scratch/sokoban/hof.txt , ver por*.txt).
Devuelve una lista con el nmero de partidas jugadas, una sublista con el numero acumulado de pasos de las soluciones encontradas y cuantas de ellas son del tipo push, y el tiempo de ejecucion total que han tardado las partidas."

  (setf *verbose-soko* verb)
  (setf fichero tipo)
  (let ((estadisticas "")
	tiempo-ejec
	(listtip (getlisttip tipo))
	listtipact
	sol
	(long-sols 0)
	(push-sols 0)
	(fallos 0)
	(exitos 0))
    (if (null num)
	(setf num (length listtip)))
    (setf tiempo-ejec (get-internal-run-time))
    (dotimes (partida num estadisticas)
      (if (null listtipact)
	  (setf listtipact listtip))
      (setf fichero (pop listtipact))
      (setf sol (funcall agente))  ;Llamada al agente
      (multiple-value-bind (long-sol push-sol) (check-soko sol) 
	(if (null long-sol)
	  (progn
	    (setf fallos (1+ fallos))
	    (format t "~%El agente fallo!!")))
	(if *verbose-soko*
	  (progn
	    (format t "~%Partida:          ~D~%" (1+ partida))
	    (format t "Long. solucion:   ~D~%" long-sol)
	    (format t "Push. solucion:   ~D~%" push-sol)
	    (format t "Solucion:         ~s~%" sol)))
	(if long-sol
	  (progn
	    (setf long-sols (+ long-sols long-sol))
	    (setf push-sols (+ push-sols push-sol))))))
    (setf tiempo-ejec (- (get-internal-run-time) tiempo-ejec))
    (setf exitos (- num fallos))
    (if *verbose-soko*
	(progn
	  (format t "~%Nmero de partidas:            ~D~%" num)
	  (format t "Nmero de xitos:              ~D~%" exitos)
	  (format t "Tipo de partidas:              ~D~%" tipo)
	  (if (plusp exitos)
	      (progn
		(format t "Promedio Long. sol.:           ~D~%" (float (/ long-sols exitos)))
		(format t "Promedio Push. sol.:           ~D~%" (float (/ push-sols exitos)))
		(format t "Tiempo promedio por partida:   ~D~%~%" (float (/ tiempo-ejec exitos)))))))
    (if (and hof (plusp exitos))
      (printhof (gen-tipo tipo)
		agente 
		(list long-sols push-sols) 
		tiempo-ejec 
		exitos))
    (list exitos (list long-sols push-sols) tiempo-ejec)))

